import os
import re

from anaconda_navigator.config import LINUX
from anaconda_navigator.utils.conda import run_process
from anaconda_navigator.utils.logs import logger


def setup_scale_factor_for_linux():
    """
        Setup QT_SCALE_FACTOR parameter to be equal to the system one,
        because QT is not able to handle scaling for the X Window Systems
    """
    if LINUX:
        os.environ["QT_SCALE_FACTOR"] = get_scaling_factor_using_dbus()


def get_scaling_factor_using_dbus():
    """Returns system primary monitor scale factor, otherwise one will be returned."""
    scaling_factor = '1'
    
    stdout, stderr, error = run_process([
        'dbus-send', '--session', '--print-reply', '--dest=org.gnome.Mutter.DisplayConfig',
        '/org/gnome/Mutter/DisplayConfig', 'org.gnome.Mutter.DisplayConfig.GetCurrentState'
    ])
    if stderr or error:
        logger.warning(
            'An exception occurred during fetching list of system display settings.')
    else:
        monitor_name = get_primary_monitor_name()
        if monitor_name:
            scaling_factor_res = re.search(
                r'struct \{\n(?:.*\n){2}.*double (?P<scaling_factor>\d(?:\.\d+)?)(?:.*\n){5}.*string \"%s\"' % monitor_name,
                stdout
            )
            if scaling_factor_res:
                scaling_factor = scaling_factor_res.group('scaling_factor')
            else:
                logger.warning('Can\'t detect system scaling factor settings for primary monitor.')

    return scaling_factor


def get_primary_monitor_name():
    """Returns name of the primary monitor"""
    primary_monitor_name = None

    stdout, stderr, error = run_process(['xrandr', '--listactivemonitors'])
    if stderr or error:
        logger.warning(
            'An exception occurred during fetching list of active monitors.')
    else:
        active_monitor_res = re.search(
            r'\d+\: \+\*(?P<primary_monitor>.*) \d*/\d+x\d+/\d+\+\d+\+\d +.*',
            stdout
        )
        if active_monitor_res:
            primary_monitor_name = active_monitor_res.group('primary_monitor')
        else:
            logger.warning('Can\'t detect primary monitor.')

    return primary_monitor_name



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
"""Preferences dialog."""

# yapf: disable

# Standard library imports
import io
import json
import os
import sys
from configparser import ConfigParser
from copy import deepcopy

# Third party imports
import yaml
from qtpy import QtCore
from qtpy.QtCore import QPoint, Qt, Signal
from qtpy.QtGui import QCursor, QPixmap
from qtpy.QtWidgets import (
    QCheckBox, QGridLayout, QHBoxLayout, QLabel,
    QLineEdit, QScrollArea, QVBoxLayout, QWidget, QTextEdit
)

# Local imports
from anaconda_navigator.api.anaconda_api import AnacondaAPI
from anaconda_navigator.api.conda_api import CondaAPI
from anaconda_navigator.config import CONF, WIN7, BITS_64, WIN
from anaconda_navigator.static.images import INFO_ICON, WARNING_ICON
from anaconda_navigator.utils.analytics import GATracker
from anaconda_navigator.widgets import (
    ButtonNormal, ButtonPrimary, SpacerHorizontal, SpacerVertical, ComboBoxBase
)
from anaconda_navigator.widgets.dialogs import DialogBase, MessageBoxError
from anaconda_navigator.widgets.dialogs.offline import DialogOfflineMode


CONDARC_DEFAULT = {
    'always_yes': True,
    'channels': ['defaults'],
    'ssl_verify': True
}


# yapf: enable


class SettingsDialog(DialogBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.text_edit = QTextEdit()
        self.text_edit.setFixedSize(600, 600)

        self.button_reset = ButtonNormal('Reset to defaults')
        self.button_reset.clicked.connect(self.reset_to_defaults)

        self.button_cancel = ButtonNormal('Cancel')
        self.button_cancel.clicked.connect(self._cancel)

        self.button_save = ButtonPrimary('Save and Restart')

        self.buttons_layout = QHBoxLayout()
        self.buttons_layout.addWidget(self.button_reset)
        self.buttons_layout.addStretch()
        self.buttons_layout.addWidget(self.button_cancel)
        self.buttons_layout.addWidget(SpacerHorizontal())
        self.buttons_layout.addWidget(self.button_save)

        self.main_layout = QVBoxLayout()
        self.main_layout.addWidget(self.text_edit)
        self.main_layout.addWidget(SpacerVertical())
        self.main_layout.addLayout(self.buttons_layout)

        self.setLayout(self.main_layout)

    def reset_to_defaults(self):
        raise NotImplementedError()

    def _cancel(self, e):
        self.close()

    @staticmethod
    def _restart_application():
        QtCore.QCoreApplication.quit()

        if WIN:
            QtCore.QProcess.startDetached(sys.argv[0])
        else:
            QtCore.QProcess.startDetached(sys.executable, sys.argv)


class NavigatorSettingsDialog(SettingsDialog):
    def __init__(self, config, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle('Navigator settings (anaconda-navigator.ini)')
        self.file_path = None
        self.config = config

        self.button_save.clicked.connect(self._save)

    def reset_to_defaults(self):
        # This hack done to have the ability to get all values
        # and update them appropriately but not touch the original
        # config, because on some widgets there are timers to write
        # data to config in some interval.
        config = deepcopy(self.config)

        self.text_edit.setText(config.get_defaults())

    def setup(self, file_path):
        if os.path.exists(file_path):
            self.file_path = file_path
            with open(file_path, 'r') as file:
                self.text_edit.setText(file.read())

    def _validate_config(self):
        """
        Validates the config data to not miss required fields
        in the config (anaconda-navigator.ini) file.
        Returns the dictionary with missing fields and sections.

        :return dict:
        """
        config_text = self.text_edit.toPlainText()

        buffer = io.StringIO(config_text)
        new_config = ConfigParser()
        new_config.read_file(buffer)

        missing_data = {}

        for section in CONF.sections():
            if not new_config.has_section(section):
                missing_data[section] = []

            for option in CONF.options(section):
                if not new_config.has_option(section, option):
                    missing_data.setdefault(section, []).append(option)

        return missing_data

    def _save(self, e):
        """
        Saves the data to the config (anaconda-navigator.ini) file if
        no missing data. Otherwise popup with missing data will arise.
        """
        missing_data = self._validate_config()

        if missing_data:
            text = (
                'The saved data missed some sections or attributes ' 
                'and could cause the issues with working Navigator! ' 
                'Please fix.'
            )
            msg_box = MessageBoxError(
                title='Navigator Settings Save Error',
                text=text,
                error=json.dumps(missing_data, indent=4),
                report=False,
                json=True
            )
            msg_box.exec_()
        else:
            if self.file_path and os.path.exists(self.file_path):
                with open(self.file_path, 'w') as file:
                    file.write(self.text_edit.toPlainText())
                    self._restart_application()


class CondaSettingsDialog(SettingsDialog):
    YAML_DOCS_URL = 'https://en.wikipedia.org/wiki/YAML'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.conda_api = CondaAPI()

        self.setWindowTitle('Conda settings (.condarc)')
        self.button_save.clicked.connect(self._save)

    def reset_to_defaults(self):
        self.text_edit.setText(yaml.dump(CONDARC_DEFAULT))

    def setup(self):
        self.text_edit.setText(self.conda_api.load_rc_plain())

    def _save(self, e):
        """
        Saves the data to config (.condarc) file if the data
        is a valid YAML format data. Otherwise the popup with
        error message will arise.
        """
        text = self.text_edit.toPlainText()

        try:
            yaml.safe_load(text)
        except yaml.YAMLError as e:
            msg_box = MessageBoxError(
                title='Conda Settings Save Error',
                text='The saved data is not a valid yaml config! Please fix.',
                error=e,
                report=False,
                learn_more=self.YAML_DOCS_URL
            )
            msg_box.exec_()
        else:
            self.conda_api.save_rc_plain(text)
            self._restart_application()


class PreferencesDialog(DialogBase):
    """Application preferences dialog."""

    sig_urls_updated = Signal(str, str)
    sig_check_ready = Signal()
    sig_reset_ready = Signal()

    def __init__(self, config=CONF, environments=None, **kwargs):
        """Application preferences dialog."""
        super(PreferencesDialog, self).__init__(**kwargs)

        self.api = AnacondaAPI()
        self.widgets_changed = set()
        self.widgets = []
        self.widgets_dic = {}
        self.config = config
        self.environments = environments
        self.tracker = GATracker()

        # Widgets
        self.button_ok = ButtonPrimary('Apply')
        self.button_cancel = ButtonNormal('Cancel')
        self.button_reset = ButtonNormal('Reset to defaults')
        self.button_nav_settings = ButtonPrimary('Configure Navigator')
        self.button_conda_settings = ButtonPrimary('Configure Conda')
        self.row = 0

        self.setFixedWidth(615)
        self.setFixedHeight(600)

        # Widget setup
        self.setWindowTitle("Preferences")

        # Layouts
        self.grid_layout = QGridLayout()

        settings_buttons_layout = QHBoxLayout()
        settings_buttons_layout.addWidget(self.button_nav_settings)
        settings_buttons_layout.addWidget(SpacerHorizontal())
        settings_buttons_layout.addWidget(self.button_conda_settings)
        settings_buttons_layout.addStretch()

        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(self.button_reset)
        buttons_layout.addStretch()
        buttons_layout.addWidget(self.button_cancel)
        buttons_layout.addWidget(SpacerHorizontal())
        buttons_layout.addWidget(self.button_ok)
        buttons_layout.setContentsMargins(0, 10, 0, 0)

        self.nav_settings = NavigatorSettingsDialog(config)
        self.conda_settings = CondaSettingsDialog()

        main_layout = QVBoxLayout()

        scroll = QScrollArea()
        scroll.setWidgetResizable(True)
        scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        scroll_widget = QWidget()
        scroll.setWidget(scroll_widget)

        content_layout = QVBoxLayout(scroll_widget)
        content_layout.addLayout(self.grid_layout)

        main_layout.addWidget(scroll)
        main_layout.addWidget(SpacerVertical())
        main_layout.addLayout(settings_buttons_layout)
        main_layout.addLayout(buttons_layout)

        self.setLayout(main_layout)

        # Signals
        self.button_ok.clicked.connect(self.accept)
        self.button_cancel.clicked.connect(self.reject)
        self.button_reset.clicked.connect(self.reset_to_defaults)
        self.button_reset.clicked.connect(
            lambda: self.button_ok.setEnabled(True)
        )
        self.button_nav_settings.clicked.connect(self.show_navigator_settings_dialog)
        self.button_conda_settings.clicked.connect(self.show_conda_settings_dialog)

        # Setup
        self.grid_layout.setSpacing(0)
        self.setup()
        self.button_ok.setDisabled(True)
        self.widgets[0].setFocus()
        self.button_ok.setDefault(True)
        self.button_ok.setAutoDefault(True)

    def show_navigator_settings_dialog(self, e):
        self.nav_settings.setup(self.config.filename())
        self.nav_settings.exec_()

    def show_conda_settings_dialog(self, e):
        self.conda_settings.setup()
        self.conda_settings.exec_()

    # --- Helpers
    # -------------------------------------------------------------------------
    def get_option(self, option):
        """Get configuration option from `main` section."""
        return self.config.get('main', option, None)

    def set_option(self, option, value):
        """Set configuration option in `main` section."""
        self.config.set('main', option, value)

    def get_option_default(self, option):
        """Get configuration option default value in `main` section."""
        return self.config.get_default('main', option)

    def set_option_default(self, option):
        """Set configuration option default value in `main` section."""
        self.set_option(option, self.get_option_default(option))

    def create_widget(
        self,
        widget=None,
        label=None,
        option=None,
        hint=None,
        check=None,
        info=None,
    ):
        """Create preference option widget and add to layout."""
        config_value = self.get_option(option)
        widget._text = label
        widget.label = QLabel(label)
        widget.option = option
        widget.set_value(config_value)
        widget.label_information = QLabel()
        widget.label_information.setMinimumWidth(16)
        widget.label_information.setMaximumWidth(16)
        widget.label_information.setMinimumHeight(16)
        widget.label_information.setMaximumHeight(16)

        form_widget = QWidget()
        h_layout = QHBoxLayout()
        h_layout.addSpacing(4)
        h_layout.addWidget(widget.label_information, 0, Qt.AlignRight)
        h_layout.addWidget(widget, 0, Qt.AlignLeft)
        h_layout.addWidget(QLabel(hint or ''), 0, Qt.AlignLeft)
        form_widget.setLayout(h_layout)

        if check:
            widget.check_value = lambda value: check(value)
        else:
            widget.check_value = lambda value: (True, '')

        if info:
            label = widget.label_information
            label = PreferencesDialog.update_icon(label, INFO_ICON)
            label.setToolTip(info)

        self.widgets.append(widget)
        self.widgets_dic[option] = widget
        self.grid_layout.addWidget(
            widget.label, self.row, 0, Qt.AlignRight | Qt.AlignCenter
        )
        self.grid_layout.addWidget(
            form_widget, self.row, 1, Qt.AlignLeft | Qt.AlignCenter
        )
        self.row += 1

    def create_textbox(self, label, option, hint=None, check=None, info=None, placeholder=None):
        """Create textbox (QLineEdit) preference option."""
        widget = QLineEdit()
        widget.setAttribute(Qt.WA_MacShowFocusRect, False)
        widget.setMinimumWidth(250)

        if placeholder:
            widget.setPlaceholderText(placeholder)

        widget.get_value = lambda w=widget: w.text()
        widget.set_value = lambda value, w=widget: w.setText(value)
        widget.set_warning = lambda w=widget: w.setSelection(0, 1000)
        widget.textChanged.connect(
            lambda v=None, w=widget: self.options_changed(widget=w)
        )

        self.create_widget(
            widget=widget,
            option=option,
            label=label,
            hint=hint,
            check=check,
            info=info,
        )

    def create_checkbox(self, label, option, check=None, hint=None, info=None):
        """Create checkbox preference option."""
        widget = QCheckBox()
        widget.get_value = lambda w=widget: bool(w.checkState())
        widget.set_value = lambda value, w=widget: bool(
            w.setCheckState(Qt.Checked if value else Qt.Unchecked)
        )

        api_widget = self.widgets_dic['anaconda_api_url']
        widget.set_warning = lambda w=widget: api_widget
        widget.stateChanged.connect(
            lambda v=None, w=widget: self.options_changed(widget=w)
        )
        self.create_widget(
            widget=widget,
            option=option,
            label=label,
            hint=hint,
            check=check,
            info=info,
        )

    def create_combobox(self, label, option, check=None, hint=None, info=None):
        widget = ComboBoxBase()
        widget.set_value = lambda *args, **kwargs: None
        widget.get_value = lambda w=widget: widget.currentData()

        default_env, selected_idx = self.get_option(option), 0

        for i, (prefix, name) in enumerate(self.environments.items()):
            if default_env == prefix:
                selected_idx = i

            widget.addItem(name, prefix)
            widget.setItemData(i, prefix, Qt.ToolTipRole)

        widget.setCurrentIndex(selected_idx)
        widget.currentIndexChanged.connect(
            lambda v=None, w=widget: self.options_changed(widget=w)
        )

        self.create_widget(
            widget=widget,
            option=option,
            label=label,
            hint=hint,
            check=check,
            info=info
        )

    def options_changed(self, value=None, widget=None):
        """Callback helper triggered on preference value change."""
        config_value = self.get_option(widget.option)

        if config_value != widget.get_value():
            self.widgets_changed.add(widget)
        else:
            if widget in self.widgets_changed:
                self.widgets_changed.remove(widget)

        self.button_ok.setDisabled(not bool(len(self.widgets_changed)))

    def widget_for_option(self, option):
        """Return the widget for the given option."""
        return self.widgets_dic[option]

    # --- API
    # -------------------------------------------------------------------------
    def set_initial_values(self):
        """
        Set configuration values found in other config files.

        Some options of configuration are found in condarc or in
        anaconda-client configuration.
        """
        # See https://conda.io/docs/install/central.html
        # ssl_verify overloads True/False/<Path to certificate>
        # Navigator splits that into 2 separate options for clarity
        ssl_verify = self.api.client_get_ssl()
        if isinstance(ssl_verify, bool):
            self.config.set('main', 'ssl_verification', ssl_verify)
            self.config.set('main', 'ssl_certificate', None)
        else:
            self.config.set('main', 'ssl_verification', True)
            self.config.set('main', 'ssl_certificate', ssl_verify)

    def setup(self):
        """Setup the preferences dialog."""

        def api_url_checker(value, allow_blank=False):
            """
            Custom checker to use selected ssl option instead of stored one.

            This allows to set an unsafe api url directly on the preferences
            dialog. Without this, one would have to first disable, click
            accept, then open preferences again and change api url for it to
            work.
            """
            # Ssl widget
            ssl_widget = self.widgets_dic.get('ssl_verification')
            verify = ssl_widget.get_value() if ssl_widget else True

            # Certificate path
            ssl_cert_widget = self.widgets_dic.get('ssl_certificate')
            if ssl_cert_widget:
                verify = ssl_cert_widget.get_value()

            # Offline mode
            offline_widget = self.widgets_dic.get('offline_mode')
            if ssl_widget or ssl_cert_widget:
                offline_mode = offline_widget.get_value()
            else:
                offline_mode = False

            if offline_mode:
                basic_check = (
                    False,
                    'API Domain cannot be modified when '
                    'working in <b>offline mode</b>.<br>',
                )
            else:
                basic_check = self.is_valid_api(value, verify=verify, allow_blank=allow_blank)

            return basic_check

        def ssl_checker(value):
            """Counterpart to api_url_checker."""
            api_url_widget = self.widgets_dic.get('anaconda_api_url')
            api_url = api_url_widget.get_value()
            return self.is_valid_api(api_url, verify=value)

        def ssl_certificate_checker(value):
            """Check if certificate path is valid/exists."""
            ssl_widget = self.widgets_dic.get('ssl_verification')
            verify = ssl_widget.get_value() if ssl_widget else True
            ssl_cert_widget = self.widgets_dic.get('ssl_certificate')
            path = ssl_cert_widget.get_value()
            return self.is_valid_cert_file(path, verify)

        self.set_initial_values()
        self.create_textbox(
            'Anaconda Cloud API domain',
            'anaconda_api_url',
            check=api_url_checker,
        )
        self.create_textbox(
            'Team Edition API domain',
            'team_edition_api_url',
            placeholder='http(s)://example.com',
            check=lambda base_url: api_url_checker(
                f'{base_url}/api/system' if base_url else None, allow_blank=True),
        )
        self.create_textbox(
            'Enterprise 4 Repository API domain',
            'enterprise_4_repo_api_url',
            placeholder='http(s)://example.com',
            check=lambda base_url: api_url_checker(base_url, allow_blank=True),
        )
        self.create_checkbox(
            'Enable SSL verification',
            'ssl_verification',
            check=ssl_checker,
            hint=(
                '<i>Disabling this option is not <br>'
                'recommended for security reasons</i>'
            ),
        )
        self.create_textbox(
            'SSL certificate path (Optional)',
            'ssl_certificate',
            check=ssl_certificate_checker,
        )

        self.create_combobox(
            'Default conda environment',
            'default_env',
        )
        info_message = '''To help us improve Anaconda Navigator, fix bugs,
and make it even easier for everyone to use Python,
we gather anonymized usage information, just like
most web browsers and mobile apps.'''
        self.create_checkbox(
            'Quality improvement reporting',
            'provide_analytics',
            info=info_message,
        )
        info_offline = DialogOfflineMode.MESSAGE_PREFERENCES
        extra = '<br><br>' if WIN7 else ''
        self.create_checkbox(
            'Enable offline mode',
            'offline_mode',
            info=info_offline + extra,
        )
        self.create_checkbox('Hide offline mode dialog', 'hide_offline_dialog')
        self.create_checkbox('Hide quit dialog', 'hide_quit_dialog')
        self.create_checkbox(
            'Hide update dialog on startup', 'hide_update_dialog'
        )
        self.create_checkbox(
            'Hide running applications dialog', 'hide_running_apps_dialog'
        )
        self.create_checkbox(
            'Enable high DPI scaling', 'enable_high_dpi_scaling'
        )
        self.create_checkbox(
            'Show application startup error messages', 'show_application_launch_errors'
        )

        ssl_ver_widget = self.widgets_dic.get('ssl_verification')
        ssl_ver_widget.stateChanged.connect(self.enable_disable_cert)
        ssl_cert_widget = self.widgets_dic.get('ssl_certificate')
        ssl_cert_widget.setPlaceholderText(
            'Certificate to verify SSL connections'
        )

        info_message = "Directory path where {0} is installed on your machine. \n" \
                       "To see {0} on the home tab, you will also need to click \n" \
                       "the Refresh button on the home tab."
        if BITS_64:
            self.create_textbox(
                'PyCharm CE path',
                'pycharm_ce_path',
                info=info_message.format("PyCharm Community")
            )

            self.create_textbox(
                'PyCharm Pro path',
                'pycharm_pro_path',
                info=info_message.format("PyCharm Professional")
            )

        self.create_textbox(
            'VS Code path',
            'vscode_path',
            info=info_message.format("Visual Studio Code")
        )

        # Refresh enabled/disabled status of certificate textbox
        self.enable_disable_cert()

    def enable_disable_cert(self, value=None):
        """Refresh enabled/disabled status of certificate textbox."""
        ssl_cert_widget = self.widgets_dic.get('ssl_certificate')
        if value:
            value = bool(value)
        else:
            ssl_ver_widget = self.widgets_dic.get('ssl_verification')
            value = bool(ssl_ver_widget.checkState())
        ssl_cert_widget.setEnabled(value)

    @staticmethod
    def update_icon(label, icon):
        """Update icon for information or warning."""
        pixmap = QPixmap(icon)
        label.setScaledContents(True)
        label.setPixmap(
            pixmap.scaled(16, 16, Qt.KeepAspectRatio, Qt.SmoothTransformation)
        )
        return label

    @staticmethod
    def warn(widget, text=None):
        """Display warning for widget in preferences."""
        label = widget.label_information
        if text:
            label = PreferencesDialog.update_icon(label, WARNING_ICON)
            label.setToolTip(str(text))
            w = widget.label_information.width() / 2
            h = widget.label_information.height() / 2
            position = widget.label_information.mapToGlobal(QPoint(w, h))
            QCursor.setPos(position)
        else:
            label.setPixmap(QPixmap())
            label.setToolTip('')

    # --- Checkers
    # -------------------------------------------------------------------------
    def is_valid_url(self, url):
        """Check if a given URL returns a 200 code."""
        output = self.api.download_is_valid_url(url, non_blocking=False)
        error = ''
        if not output:
            error = 'Invalid api url.'
        return output, error

    def is_valid_cert_file(self, path, verify):
        """"Check if ssl certificate file in given path exists."""
        output = True
        error = ''

        # Only validate if it is not empty and if ssl_verification is checked
        if path.strip() and verify:
            output = os.path.isfile(path)
            if not output:
                error = 'File not found.'
        return output, error

    def is_valid_api(self, url, verify=True, allow_blank=False):
        """Check if a given URL is a valid anaconda api endpoint."""
        output = self.api.download_is_valid_api_url(
            url,
            non_blocking=False,
            verify=verify,
            allow_blank=allow_blank
        )
        error = ''
        if not output:
            url_api_1 = ''
            url_api_2 = ''

            if '/api' not in url and self.is_valid_url(url)[0]:
                url_api_1 = url.replace('https://', 'https://api.')
                url_api_1 = url_api_1.replace('http://', 'http://api.')

                if url.endswith('/'):
                    url_api_2 = url + 'api'
                else:
                    url_api_2 = url + '/api'

                error = (
                    'Invalid API url. <br><br>'
                    '<br>Try using:<br><b>{0}</b> or <br>'
                    '<b>{1}</b>'.format(url_api_1, url_api_2)
                )
            else:
                error = (
                    'Invalid API url. <br><br>'
                    'Check the url is valid and corresponds to the api '
                    'endpoint.'
                )
        return output, error

    def run_checks(self):
        """
        Run all check functions on configuration options.

        This method checks and warns but it does not change/set values.
        """
        checks = []
        for widget in self.widgets_changed:
            value = widget.get_value()
            check, error = widget.check_value(value)
            checks.append(check)

            if check:
                self.warn(widget)
            else:
                self.button_ok.setDisabled(True)
                widget.set_warning()
                self.warn(widget, error)
                break

        # Emit checks ready
        self.sig_check_ready.emit()
        return checks

    def reset_to_defaults(self):
        """Reset the preferences to the default values."""
        for widget in self.widgets:
            option = widget.option
            default = self.get_option_default(option)
            widget.set_value(default)

            # Flag all values as updated
            self.options_changed(widget=widget, value=default)
        self.sig_reset_ready.emit()

    def accept(self):
        """Override Qt method."""
        sig_updated = False
        anaconda_api_url = None
        checks = self.run_checks()

        # Update values
        if checks and all(checks):
            for widget in self.widgets_changed:
                value = widget.get_value()
                self.set_option(widget.option, value)

                if widget.option == 'default_env':
                    self.tracker.track_event('preferences', 'change', 'Default environment')

                # Settings not stored on Navigator config, but taken from
                # anaconda-client config
                if widget.option == 'anaconda_api_url':
                    anaconda_api_url = value  # Store it to be emitted
                    self.api.client_set_api_url(value)
                    sig_updated = True

                # ssl_verify/verify_ssl handles True/False/<Path to cert>
                # On navi it is split in 2 options for clarity
                if widget.option in ['ssl_certificate', 'ssl_verification']:
                    ssl_veri = self.widgets_dic.get('ssl_verification')
                    ssl_cert = self.widgets_dic.get('ssl_certificate')
                    verify = ssl_veri.get_value()
                    path = ssl_cert.get_value()

                    if path.strip() and verify:
                        value = path
                    else:
                        value = verify

                    self.api.client_set_ssl(value)

            if sig_updated and anaconda_api_url:

                def _api_info(worker, output, error):
                    conda_url = output.get('conda_url')
                    try:
                        self.sig_urls_updated.emit(anaconda_api_url, conda_url)
                        super(PreferencesDialog, self).accept()
                    except RuntimeError:
                        # Some tests on appveyor/circleci fail
                        pass

                worker = self.api.api_urls()
                worker.sig_chain_finished.connect(_api_info)

            super(PreferencesDialog, self).accept()


# --- Local testing
# -----------------------------------------------------------------------------
def local_test():  # pragma: no cover
    """Main local testing."""
    from anaconda_navigator.utils.qthelpers import qapplication
    app = qapplication()
    widget = PreferencesDialog(parent=None)
    widget.show()
    app.exec_()


if __name__ == '__main__':  # pragma: no cover
    local_test()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
"""Splash screen and intial startup splash."""

# yapf: disable

from __future__ import absolute_import, division, print_function

# Third party imports
from qtpy.QtCore import Qt
from qtpy.QtGui import QColor, QPixmap
from qtpy.QtWidgets import QApplication, QGraphicsOpacityEffect, QSplashScreen

# Local imports
from anaconda_navigator.static.images import ANACONDA_ICON_256_PATH


# yapf: enable


class SplashScreen(QSplashScreen):
    """Splash screen for the main window."""

    def __init__(self, *args, **kwargs):
        """Splash screen for the main window."""
        super(SplashScreen, self).__init__(*args, **kwargs)
        self._effect = QGraphicsOpacityEffect()
        self._font = self.font()
        self._pixmap = QPixmap(ANACONDA_ICON_256_PATH)
        self._message = ''

        # Setup
        self._font.setPixelSize(10)
        self._effect.setOpacity(0.9)
        self.setFont(self._font)
        self.setGraphicsEffect(self._effect)
        self.setPixmap(self._pixmap)
        self.setWindowFlags(
            Qt.FramelessWindowHint | Qt.SplashScreen | Qt.WindowStaysOnTopHint
        )

    def get_message(self):
        """Return currently displayed message."""
        return self._message

    def show_message(self, message):
        """Show message in the screen."""
        self._message = message
        message += '\n'
        self.show()
        self.showMessage(
            message, Qt.AlignBottom | Qt.AlignCenter | Qt.AlignAbsolute,
            QColor(Qt.white)
        )
        QApplication.processEvents()


# --- Local testing
# -----------------------------------------------------------------------------
def local_test():  # pragma: no cover
    """Run local test."""
    from anaconda_navigator.utils.qthelpers import qapplication

    app = qapplication()
    widget_splash = SplashScreen()
    widget_splash.show_message('Initializing...')
    widget_splash.show()

    app.exec_()


if __name__ == '__main__':  # pragma: no cover
    local_test()



from anaconda_navigator.api.external_apps.base import BaseWebApp
from anaconda_navigator.static import images


class IBMWatsonApp(BaseWebApp):
    def __init__(self, **kwargs):
        super(IBMWatsonApp, self).__init__(
            app_name='ibm_watson',
            display_name='IBM Watson Studio Cloud',
            description='IBM Watson Studio Cloud provides you the tools to analyze and visualize data, '
                        'to cleanse and shape data, to create and train machine learning models. Prepare data and '
                        'build models, using open source data science tools or '
                        'visual modeling.',
            image_path=images.IBM_WATSON_ICON_PATH,
            needs_license=False,
            url='http://www.anaconda.com/ibm_wsc_navigator',
            **kwargs
        )


import ctypes
import datetime
import io
import json
import os

from anaconda_navigator.api.external_apps.base import BaseInstallableApp
from anaconda_navigator.config import CONF_PATH, LINUX, MAC, BITS_32
from anaconda_navigator.static import images
from anaconda_navigator.utils.conda import run_process
from anaconda_navigator.utils.logs import logger
from anaconda_navigator.utils.py3compat import PY3


class VSCodeApp(BaseInstallableApp):
    def __init__(self, **kwargs):
        super(VSCodeApp, self).__init__(
            app_name='vscode',
            filename='code',
            windows_folder_name='Microsoft VS Code',
            mac_name='Visual Studio Code.app',
            display_name='VS Code',
            description=('Streamlined code editor with support for '
                         'development operations like debugging, task '
                         'running and version control.'),
            image_path=images.VSCODE_ICON_1024_PATH,
            needs_license=False,
            is_installation_enabled=False,
            **kwargs
        )

        self._vscode_version_value = None

    @property
    def executable(self) -> str:
        """Return an executable command to launch the app."""
        if MAC:
            executable = os.path.join(
                self.app_directory_path, 'Contents', 'Resources', 'app', 'bin', self.filename
            )
        elif LINUX:
            if 'share' in self.app_directory_path:
                executable = os.path.join(self.app_directory_path, 'bin', self.filename)
            else:
                executable = os.path.join(
                    self.app_directory_path, 'usr', 'share', 'code', 'bin', self.filename
                )
        else:
            executable = os.path.join(self.app_directory_path, 'bin', f'{self.filename}.cmd')

        if os.path.exists(executable):
            return f'"{executable}"'

        return ''

    def _get_linux_installation_directory(self) -> str:
        root_dirs = ['/usr/share/']

        if os.path.exists('/snap'):
            root_dirs.append('/snap')

        for root in root_dirs:
            if os.path.exists(os.path.join(root, self.filename)):
                # using Snap user can install several versions, 'current' is a link to newest
                return os.path.join(root, self.filename, 'current' if root == '/snap' else '')
        return ''

    def _get_windows_installation_directory(self) -> str:
        from anaconda_navigator.external.knownfolders import get_folder_path, FOLDERID

        _kernel32 = ctypes.windll.kernel32
        _windir = ctypes.create_unicode_buffer(1024)
        _kernel32.GetWindowsDirectoryW(_windir, 1024)
        _windrive = _windir.value[:3]

        if BITS_32:
            program_files_dir = get_folder_path(FOLDERID.ProgramFilesX86)[0]
        else:
            program_files_dir = get_folder_path(FOLDERID.ProgramFilesX64)[0]

        local_data_dir = get_folder_path(FOLDERID.LocalAppData)[0]
        program_files_64_dir = (
                get_folder_path(FOLDERID.ProgramFilesX64)[0]
                or os.path.join(_windrive, 'Program Files')
        )

        if os.path.exists(os.path.join(program_files_dir, self.windows_folder_name)):
            application_path = os.path.join(program_files_dir, self.windows_folder_name)
        elif os.path.exists(os.path.join(program_files_64_dir, self.windows_folder_name)):
            application_path = os.path.join(program_files_64_dir, self.windows_folder_name)
        else:
            application_path = os.path.join(local_data_dir, self.windows_folder_name)
            if not os.path.isdir(application_path):
                application_path = os.path.join(local_data_dir, 'Programs', self.windows_folder_name)

        return application_path

    @property
    def versions(self):
        if self._vscode_version_value is None and self.is_installed:
            stdout, _, _ = run_process([self.executable[1:-1], '--version'])
            # self.executable[1:-1] to remove unnecessary ""
            if stdout:
                version = stdout.split('\n')[0]
                self._vscode_version_value = version
        return [self._vscode_version_value]

    def install_extensions(self):
        """Install app extensions."""
        wm = self._process_api
        logger.debug('Installing vscode extensions')
        cmd = [
            '{}'.format(self.executable),
            '--install-extension',
            'ms-python.anaconda-extension-pack',
            # ms-python-anaconda-extension
            # ms-python.python
        ]
        logger.debug(' '.join(cmd))
        worker = wm.create_process_worker(cmd)
        return worker

    def update_config(self, prefix):
        logger.debug('Update app config to use prefix {}'.format(prefix))
        try:
            _config = os.path.join(
                CONF_PATH,
                'Code',
                'User',
                'settings.json',
            )
            _config_dir = os.path.dirname(_config)

            try:
                if not os.path.isdir(_config_dir):
                    os.makedirs(_config_dir)
            except Exception as e:
                logger.error(e)

            config_update = {'python.pythonPath': prefix}

            if os.path.isfile(_config):
                try:
                    with io.open(_config, 'r', encoding='utf-8') as f:
                        data = f.read()

                    self.create_config_backup(data)

                    config_data = json.loads(data)
                    for key, val in config_update.items():
                        config_data[key] = val
                except Exception:
                    # If there is any error, don't overwrite app config
                    return False
            else:
                config_data = config_update.copy()

            mode = 'w' if PY3 else 'wb'
            with io.open(_config, mode) as f:
                json.dump(
                    config_data,
                    f,
                    sort_keys=True,
                    indent=4,
                )
        except Exception as e:
            logger.error(e)
            return False

        return True

    def create_config_backup(self, data):
        """
        Create a backup copy of the app configuration file `data`.

        Leave only the last 10 backups.
        """
        now = datetime.datetime.now()
        date = now.strftime('%Y%m%d%H%M%S')
        _config_dir = os.path.join(
            CONF_PATH,
            'Code',
            'User',
        )
        _config_bck = os.path.join(
            _config_dir,
            'bck.{date}.navigator.settings.json'.format(date=date),
        )

        # Make the backup
        logger.debug(
            'Creating backup app config file: {}'
            ''.format(_config_bck)
        )
        with io.open(_config_bck, 'w', encoding='utf-8') as f_handle:
            f_handle.write(data)

        # Only keep the latests 10 backups
        files = os.listdir(_config_dir)
        fpaths = [
            os.path.join(_config_dir, f) for f in files
            if f.startswith('bck.') and f.endswith('.navigator.settings.json')
        ]

        fpaths_remove = list(sorted(fpaths, reverse=True))[10:]
        for fpath in fpaths_remove:
            try:
                os.remove(fpath)
            except Exception:
                pass



import requests
from http import HTTPStatus
from urllib.parse import urlparse

from anaconda_navigator.utils.logs import logger


class AdvertisementMixin:
    def _is_valid_url(self, url):
        parse_result = urlparse(url)

        if not all((parse_result.scheme, parse_result.netloc, parse_result.path)):
            self.errors.append('{} is not a valid URL'.format(url))
            return False
        return True
    
    def _log_errors(self):
        for error in self.errors:
            logger.error(error)


class AdvertisementsList(AdvertisementMixin):
    BANNER_TIMEOUT_MS = 60000

    def __init__(self, source_url):
        self.source_url = source_url
        self.advertisements = []
        self.errors = []
        self.load()

        self._log_errors()
    
    def load(self):
        """
        Tries to load the data to be displayed in the advertisements.
        """
        try:
            resp = requests.get(self.source_url)

            if resp.status_code == HTTPStatus.OK:
                for data in resp.json():
                    image_url = data.get('image_url')
                    redirect_url = data.get('redirect_url')
                    text = data.get('text')
                    
                    ad = Advertisement(image_url, redirect_url, text)

                    if ad.is_valid():
                        self.advertisements.append(ad)

        except requests.ConnectionError:
            self.errors.append(
                'Tried to load advertisement data from {}. URL is not reachable'.format(self.source_url)
            )
        except Exception as e:
            self.errors.append(
                'Unexpected error while loading advertisement. Stacktrace: {}'.format(str(e))
            )

    def advertisements_carousel(self):
        start_pos = 0
        adlen = len(self.advertisements)
        while start_pos < adlen:
            curr_pos = start_pos
            res = self.advertisements[start_pos]
            start_pos = (start_pos + 1) if start_pos + 1 < adlen else 0
            yield curr_pos, res
        

class Advertisement(AdvertisementMixin):
    """
    Class for loading advertisements data, handling errors and validating data.
    """

    def __init__(self, image_url, redirect_url, text):
        self.errors = []
        self.image = None
        if self._is_valid_url(image_url):
            self.image = self._load_image(image_url)
        
        self.url = None
        if self._is_valid_url(redirect_url):
            self.url = redirect_url

        self.text = text

        self._log_errors()

    def is_valid(self):
        """
        Checks if all needed data is present and correct.
        """
        image_and_link = self.image and self.url
        text_and_link = self.text and self.url
        all_set = self.text and self.url and self.image

        return all_set or image_and_link or text_and_link

    def _load_image(self, image_url):
        """
        Tries to load the image of an advertisement.

        :param str image_url: The url of the image to be downloaded.
        """
        resp = requests.get(image_url)
        
        if resp.status_code == HTTPStatus.OK:
            return resp.content
        else:
            self.errors.append(
                'Tried to load advertisement image from {}. Data is not fetched with status {}'.format(
                    image_url, resp.status_code
                )
            )



from anaconda_navigator.api.external_apps.base import BaseWebApp
from anaconda_navigator.static import images


class DataloreApp(BaseWebApp):
    def __init__(self, **kwargs):
        super(DataloreApp, self).__init__(
            app_name='datalore',
            display_name='Datalore',
            description='Online Data Analysis Tool with smart coding assistance by JetBrains. '
                        'Edit and run your Python notebooks in the cloud and share them with your team.',
            image_path=images.DATALORE_ICON_1024_PATH,
            needs_license=False,
            url='http://www.anaconda.com/datalore_navigator',
            **kwargs
        )



import os
import webbrowser
from abc import ABC, abstractmethod

from anaconda_navigator.config import MAC, LINUX, WIN
from anaconda_navigator.utils.constants import AppType


class BaseApp(ABC):
    def __init__(self, config, app_name, display_name, description, image_path, needs_license, non_conda, app_type):
        self.config = config
        self.app_name = app_name
        self.display_name = display_name
        self.description = description
        self.image_path = image_path
        self.needs_license = needs_license
        self.non_conda = non_conda
        self.app_type = app_type
        self.is_available = True


class BaseWebApp(BaseApp, ABC):
    def __init__(self, url, **kwargs):
        super(BaseWebApp, self).__init__(app_type=AppType.WEB, non_conda=True, **kwargs)

        self.url = url

    def launch(self):
        """Launch the application."""
        webbrowser.open_new_tab(self.url)


class BaseInstallableApp(BaseApp, ABC):
    def __init__(
            self, filename, windows_folder_name, mac_name, is_installation_enabled, process_api, **kwargs
    ):
        super(BaseInstallableApp, self).__init__(app_type=AppType.INSTALLABLE, non_conda=True, **kwargs)

        self.filename = filename
        self.windows_folder_name = windows_folder_name
        self.mac_name = mac_name
        self.is_installation_enabled = is_installation_enabled
        self.app_directory_path = ''

        self._process_api = process_api

        self.set_up_app_directory_path()

    def set_up_app_directory_path(self):
        """Set up valid app_directory_path."""
        config_app_directory_path = self.config.get('main', f'{self.app_name}_path', '')

        if config_app_directory_path and os.path.exists(config_app_directory_path):
            self.app_directory_path = config_app_directory_path
        else:
            if MAC:
                self.app_directory_path = self._get_macos_installation_directory()
            elif LINUX:
                self.app_directory_path = self._get_linux_installation_directory()
            elif WIN:
                self.app_directory_path = self._get_windows_installation_directory()

            self.config.set('main', f'{self.app_name}_path', self.app_directory_path)

    @property
    @abstractmethod
    def executable(self):
        """Return executable command."""
        raise NotImplementedError

    @property
    def is_installed(self):
        """Return whether the app is installed."""
        return bool(self.executable)

    def _get_macos_installation_directory(self) -> str:
        for root in ('/Applications', os.path.join(os.path.expanduser('~'), 'Applications')):
            if os.path.exists(os.path.join(root, self.mac_name)):
                return os.path.join(root, self.mac_name)
        return ''

    @abstractmethod
    def _get_linux_installation_directory(self):
        raise NotImplementedError

    @abstractmethod
    def _get_windows_installation_directory(self):
        raise NotImplementedError

    @abstractmethod
    def update_config(self, prefix):
        """Update user config to use selected Python prefix interpreter."""
        raise NotImplementedError

    @abstractmethod
    def install_extensions(self):
        """Install app extensions."""
        raise NotImplementedError

    def install(self):
        """Install app."""
        # this method should be implement only if self.install_enabled == True
        raise NotImplementedError



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
"""
Home Tab.

This widget does not perform the actual actions but it emits signals that
should be connected to the final controller on the main window.
"""

# yapf: disable

from __future__ import absolute_import, division, print_function

# Standard library imports
import sys

# Third party imports
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import (QApplication, QHBoxLayout, QLabel, QProgressBar,
                            QVBoxLayout)

# Local imports
from anaconda_navigator.api.anaconda_api import AnacondaAPI
from anaconda_navigator.utils import constants as C
from anaconda_navigator.utils.analytics import GATracker
from anaconda_navigator.utils.styles import load_style_sheet
from anaconda_navigator.widgets import (ButtonNormal, ComboBoxBase,
                                        FrameTabContent, FrameTabFooter,
                                        FrameTabHeader, LabelBase,
                                        SpacerHorizontal, WidgetBase)
from anaconda_navigator.widgets.lists.apps import (ListItemApplication,
                                                   ListWidgetApplication)


# yapf: enable


# --- Custom widgets used with CSS styling
# -----------------------------------------------------------------------------
class ButtonHomeRefresh(ButtonNormal):
    """QFrame used for CSS styling refresh button inside the Home Tab."""


class ComboHomeEnvironment(ComboBoxBase):
    """Widget Used for CSS styling."""


class ButtonHomeChannels(ButtonNormal):
    """Widget Used for CSS styling."""


class LabelHome(LabelBase):
    """QLabel used for CSS styling the Home Tab label."""


# --- Main widget
# -----------------------------------------------------------------------------
class HomeTab(WidgetBase):
    """Home applications tab."""
    # name, prefix, sender
    sig_item_selected = Signal(object, object, object)

    # button_widget, sender
    sig_channels_requested = Signal(object, object)

    # application_name, command, prefix, leave_path_alone, sender, non_conda, app_type
    sig_launch_action_requested = Signal(
        object, object, bool, object, object, object, object
    )

    # action, application_name, version, sender, non_conda, app_type
    sig_conda_action_requested = Signal(object, object, object, object, object, object)

    # url
    sig_url_clicked = Signal(object)

    # TODO: Connect these signals to have more granularity
    # [{'name': package_name, 'version': version}...], sender
    sig_install_action_requested = Signal(object, object)
    sig_remove_action_requested = Signal(object, object)

    def __init__(self, parent=None):
        """Home applications tab."""
        super(HomeTab, self).__init__(parent)

        # Variables
        self._parent = parent
        self.api = AnacondaAPI()
        self.applications = None
        self.style_sheet = None
        self.app_timers = None
        self.current_prefix = None
        self.tracker = GATracker()

        # Widgets
        self.list = ListWidgetApplication()
        self.button_channels = ButtonHomeChannels('Channels')
        self.button_refresh = ButtonHomeRefresh('Refresh')
        self.combo = ComboHomeEnvironment()
        self.frame_top = FrameTabHeader(self)
        self.frame_body = FrameTabContent(self)
        self.frame_bottom = FrameTabFooter(self)
        self.label_home = LabelHome('Applications on')
        self.label_status_action = QLabel('')
        self.label_status = QLabel('')
        self.progress_bar = QProgressBar()
        self.first_widget = self.combo

        # Widget setup
        self.setObjectName('Tab')
        self.progress_bar.setTextVisible(False)
        self.list.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)

        # Layout
        layout_top = QHBoxLayout()
        layout_top.addWidget(self.label_home)
        layout_top.addWidget(SpacerHorizontal())
        layout_top.addWidget(self.combo)
        layout_top.addWidget(SpacerHorizontal())
        layout_top.addWidget(self.button_channels)
        layout_top.addWidget(SpacerHorizontal())
        layout_top.addStretch()
        layout_top.addWidget(self.button_refresh)
        self.frame_top.setLayout(layout_top)

        layout_body = QVBoxLayout()
        layout_body.addWidget(self.list)
        self.frame_body.setLayout(layout_body)

        layout_bottom = QHBoxLayout()
        layout_bottom.addWidget(self.label_status_action)
        layout_bottom.addWidget(SpacerHorizontal())
        layout_bottom.addWidget(self.label_status)
        layout_bottom.addStretch()
        layout_bottom.addWidget(self.progress_bar)
        self.frame_bottom.setLayout(layout_bottom)

        layout = QVBoxLayout()
        layout.addWidget(self.frame_top)
        layout.addWidget(self.frame_body)
        layout.addWidget(self.frame_bottom)
        self.setLayout(layout)

        # Signals
        self.list.sig_conda_action_requested.connect(
            self.sig_conda_action_requested
        )
        self.list.sig_url_clicked.connect(self.sig_url_clicked)
        self.list.sig_launch_action_requested.connect(
            self.sig_launch_action_requested
        )
        self.button_channels.clicked.connect(self.show_channels)
        self.button_refresh.clicked.connect(self.refresh_cards)
        self.progress_bar.setVisible(False)

    # --- Setup methods
    # -------------------------------------------------------------------------
    def setup(self, conda_data):
        """Setup the tab content."""
        conda_processed_info = conda_data.get('processed_info')
        environments = conda_processed_info.get('__environments')
        applications = conda_data.get('applications')
        packages = conda_data.get('packages')
        self.current_prefix = conda_processed_info.get('default_prefix')
        self.set_environments(environments)
        self.set_applications(applications, packages)

    def set_environments(self, environments):
        """Setup the environments list."""
        # Disconnect to avoid triggering the signal when updating the content
        try:
            self.combo.currentIndexChanged.disconnect()
        except TypeError:
            pass

        self.combo.clear()

        fm = self.combo.fontMetrics()
        widths = []

        for i, (env_prefix, env_name) in enumerate(environments.items()):
            widths.append(fm.width(env_name))
            self.combo.addItem(env_name, env_prefix)
            self.combo.setItemData(i, env_prefix, Qt.ToolTipRole)

        index = 0
        for i, (env_prefix, env_name) in enumerate(environments.items()):
            if self.current_prefix == env_prefix:
                index = i
                break

        self.combo.setCurrentIndex(index)
        self.combo.currentIndexChanged.connect(self._item_selected)

        # Fix combobox width
        width = max(widths) + 64
        self.combo.setMinimumWidth(width)

    def set_applications(self, applications, packages):
        """Build the list of applications present in the current conda env."""
        apps = self.api.process_apps(applications, prefix=self.current_prefix)
        all_applications = []
        installed_applications = []
        not_installed_applications = []

        # Check if some installed applications are not on the apps dict
        # for example when the channel was removed.
        linked_apps = self.api.conda_linked_apps_info(self.current_prefix)
        missing_apps = [app for app in linked_apps if app not in apps]
        for app in missing_apps:
            apps[app] = linked_apps[app]

        for app_name in sorted(list(apps.keys())):
            app = apps[app_name]
            item = ListItemApplication(prefix=self.current_prefix, **app)
            if app.get('installed'):
                installed_applications.append(item)
            else:
                not_installed_applications.append(item)

        all_applications = installed_applications + not_installed_applications

        self.list.clear()
        for i in all_applications:
            self.list.addItem(i)
        self.list.update_style_sheet(self.style_sheet)

        self.set_widgets_enabled(True)
        self.update_status()

    # --- Other methods
    # -------------------------------------------------------------------------
    def current_environment(self):
        """Return the current selected environment."""
        env_name = self.combo.currentText()
        return self.api.conda_get_prefix_envname(env_name)

    def refresh_cards(self):
        """Refresh application widgets.

        List widget items sometimes are hidden on resize. This method tries
        to compensate for that refreshing and repainting on user demand.
        """
        self.tracker.track_event('home', 'click', 'Refresh')
        self.list.update_style_sheet(self.style_sheet)
        self.list.repaint()
        for item in self.list.items():
            if not item.widget.isVisible():
                item.widget.repaint()

        worker = self.api.conda_data(prefix=self.current_prefix)
        worker.sig_chain_finished.connect(lambda _, output, __: self.setup(output))
        self.update_status(action=f'Refreshing applications', value=0, max_value=0)

    def show_channels(self):
        """Emit signal requesting the channels dialog editor."""
        self.tracker.track_event('home', 'click', 'Channels')
        self.sig_channels_requested.emit(self.button_channels, C.TAB_HOME)

    def update_versions(self, apps=None):
        """Update applications versions."""
        self.items = []

        for i in range(self.list.count()):
            item = self.list.item(i)
            self.items.append(item)
            if isinstance(item, ListItemApplication):
                name = item.name
                meta = apps.get(name)
                if meta:
                    versions = meta['versions']
                    version = self.api.get_dev_tool_version(item.path)
                    item.update_versions(version, versions)

    # --- Common Helpers (# FIXME: factor out to common base widget)
    # -------------------------------------------------------------------------
    def _item_selected(self, index):
        """Notify that the item in combo (environment) changed."""
        name = self.combo.itemText(index)
        prefix = self.combo.itemData(index)
        self.sig_item_selected.emit(name, prefix, C.TAB_HOME)

    @property
    def last_widget(self):
        """Return the last element of the list to be used in tab ordering."""
        if self.list.items():
            return self.list.items()[-1].widget

    def ordered_widgets(self, next_widget=None):
        """Return a list of the ordered widgets."""
        ordered_widgets = [
            self.combo,
            self.button_channels,
            self.button_refresh,
        ]
        ordered_widgets += self.list.ordered_widgets()

        return ordered_widgets

    def set_widgets_enabled(self, value):
        """Enable or disable widgets."""
        self.combo.setEnabled(value)
        self.button_channels.setEnabled(value)
        self.button_refresh.setEnabled(value)
        for item in self.list.items():
            item.button_install.setEnabled(value)
            item.button_options.setEnabled(value)

            if value:
                item.set_loading(not value)

    def update_items(self):
        """Update status of items in list."""
        if self.list:
            for item in self.list.items():
                item.update_status()

    def update_status(self, action='', message='', value=None, max_value=None):
        """Update the application action status."""

        # Elide if too big
        width = QApplication.desktop().availableGeometry().width()
        max_status_length = round(width * (2.0 / 3.0), 0)
        msg_percent = 0.70

        fm = self.label_status_action.fontMetrics()
        action = fm.elidedText(
            action, Qt.ElideRight, round(max_status_length * msg_percent, 0)
        )
        message = fm.elidedText(
            message, Qt.ElideRight,
            round(max_status_length * (1 - msg_percent), 0)
        )
        self.label_status_action.setText(action)
        self.label_status.setText(message)

        if max_value is None and value is None:
            self.progress_bar.setVisible(False)
        else:
            self.progress_bar.setVisible(True)
            self.progress_bar.setMaximum(max_value)
            self.progress_bar.setValue(value)

    def update_style_sheet(self, style_sheet=None):
        """Update custom CSS style sheet."""
        if style_sheet is None:
            self.style_sheet = load_style_sheet()
        else:
            self.style_sheet = style_sheet

        self.list.update_style_sheet(style_sheet=self.style_sheet)
        self.setStyleSheet(self.style_sheet)


# --- Local testing
# -----------------------------------------------------------------------------
def local_test():  # pragma: no cover
    """Run local test."""
    from anaconda_navigator.utils.qthelpers import qapplication

    app = qapplication()
    widget = HomeTab()
    apps = {}
    for i in range(6):
        name = "Package {0}".format(i)
        application = dict(
            name=name,
            prefix='boom',
            description="Scientific PYthon Development EnviRonment",
            versions=[str(i), str(i + 1), str(i + 2)],
            command=None,
            image_path=None
        )
        apps[name] = application
    widget.setup(apps)
    widget.update_style_sheet()
    widget.showMaximized()
    sys.exit(app.exec_())


if __name__ == "__main__":  # pragma: no cover
    local_test()



from .datalore import DataloreApp
from .ibm_watson import IBMWatsonApp
from .pycharm import PyCharmCEApp, PyCharmProApp
from .vscode import VSCodeApp

apps = {'vscode': VSCodeApp, 'pycharm_pro': PyCharmProApp, 'pycharm_ce': PyCharmCEApp}
web_apps = {'datalore': DataloreApp, 'ibm_watson': IBMWatsonApp}



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
"""Anaconda Cloud login dialog."""

# yapf: disable

from __future__ import absolute_import, division, print_function

# Standard library imports
import ast
import json

# Third party imports
from qtpy.QtCore import QRegExp, Qt, QUrl, Signal
from qtpy.QtGui import QDesktopServices, QRegExpValidator
from qtpy.QtWidgets import (QApplication, QHBoxLayout, QLabel, QLineEdit,
                            QVBoxLayout, QWidget, QTabWidget)

# Local imports
from anaconda_navigator.api.anaconda_api import AnacondaAPI
from anaconda_navigator.api.team_edition_api import TeamEditionAPI
from anaconda_navigator.config import (
    CONF, DEFAULT_BRAND, ENTERPRISE_BRAND, TEAM_EDITION_BRAND
)
from anaconda_navigator.static.fonts import load_fonts
from anaconda_navigator.utils.analytics import GATracker
from anaconda_navigator.utils.styles import load_style_sheet
from anaconda_navigator.widgets import (ButtonLabel, ButtonLink, ButtonNormal,
                                        ButtonPrimary, SpacerHorizontal,
                                        SpacerVertical)
from anaconda_navigator.widgets.dialogs import DialogBase, MessageBoxInformation


# yapf: enable

USER_RE = QRegExp(r'^[A-Za-z0-9_\.][A-Za-z0-9_\.\-]+$')


class LoginTabMixin(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def username(self):
        """Return the logged username."""
        return self.text_username.text().lower()

    def check_text(self):
        """Check that `username` and `password` are not empty strings.

        If not empty and disable/enable buttons accordingly.
        """
        username = self.text_username.text()
        password = self.text_password.text()

        if not all((len(username), len(password))):
            self.button_login.setDisabled(True)
        else:
            self.button_login.setDisabled(False)

    def update_style_sheet(self, style_sheet=None):
        """Update custom css style sheet."""
        if style_sheet is None:
            style_sheet = load_style_sheet()
        self.setStyleSheet(style_sheet)

    def reject(self):
        self._parent.reject()


class AnacondaCloudLoginTab(LoginTabMixin):

    # See https://github.com/Anaconda-Platform/anaconda-server settings
    FORGOT_USERNAME_URL = 'account/forgot_username'
    FORGOT_PASSWORD_URL = 'account/forgot_password'

    sig_url_clicked = Signal(object)
    sig_authentication_succeeded = Signal()
    sig_authentication_failed = Signal()

    def __init__(self, anaconda_api, parent, brand, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._brand = brand
        self._parent = parent
        self.config = CONF
        self.anaconda_api = anaconda_api
        self.token = None
        self.error = None
        self.tracker = GATracker()
        self.forgot_username_url = None
        self.forgot_password_url = None

        # Widgets
        self.label_username = QLabel('Username:')
        self.label_password = QLabel('Password:')
        self.text_username = QLineEdit()
        self.text_password = QLineEdit()
        self.label_signin_text = QLabel(
            '<hr><br><b>Already a member? '
            'Sign in!</b><br>'
        )

        # For styling purposes the label next to a ButtonLink is also a button
        # so they align adequately
        self.button_register_text = ButtonLabel(
            'You can register by '
            'visiting the '
        )
        self.button_register = ButtonLink(brand)
        self.button_register_after_text = ButtonLabel(' website.')
        self.label_information = QLabel(
            '''
            <strong>{brand}</strong> is where packages, notebooks,
            and <br> environments are shared. It provides powerful <br>
            collaboration and package management for open <br>
            source and private projects.<br>
            '''.format(brand=brand)
        )
        self.label_message = QLabel('')
        self.label_message.setWordWrap(True)
        self.button_forgot_username = ButtonLink('I forgot my username')
        self.button_forgot_password = ButtonLink('I forgot my password')
        self.button_login = ButtonPrimary('Login')
        self.button_cancel = ButtonNormal('Cancel')

        # Widgets setup
        self.button_login.setDefault(True)
        username_validator = QRegExpValidator(USER_RE)
        self.text_username.setValidator(username_validator)

        # This allows to completely style the dialog with css using the frame
        self.text_password.setEchoMode(QLineEdit.Password)
        self.label_message.setVisible(False)

        # Layout
        grid_layout = QVBoxLayout()
        grid_layout.addWidget(self.label_username)
        grid_layout.addWidget(self.text_username)
        grid_layout.addWidget(SpacerVertical())
        grid_layout.addWidget(self.label_password)
        grid_layout.addWidget(self.text_password)

        main_layout = QVBoxLayout()
        main_layout.addWidget(self.label_information)

        register_layout = QHBoxLayout()
        register_layout.addWidget(self.button_register_text)
        register_layout.addWidget(self.button_register)
        register_layout.addWidget(self.button_register_after_text)
        register_layout.addStretch()
        main_layout.addLayout(register_layout)
        main_layout.addWidget(self.label_signin_text)
        main_layout.addLayout(grid_layout)
        main_layout.addWidget(SpacerVertical())
        main_layout.addWidget(self.label_message)
        main_layout.addWidget(self.button_forgot_username, 0, Qt.AlignRight)
        main_layout.addWidget(self.button_forgot_password, 0, Qt.AlignRight)

        layout_buttons = QHBoxLayout()
        layout_buttons.addStretch()
        layout_buttons.addWidget(self.button_cancel)
        layout_buttons.addWidget(SpacerHorizontal())
        layout_buttons.addWidget(self.button_login)

        main_layout.addWidget(SpacerVertical())
        main_layout.addWidget(SpacerVertical())
        main_layout.addLayout(layout_buttons)

        # Signals
        self.text_username.textEdited.connect(self.check_text)
        self.text_password.textEdited.connect(self.check_text)
        self.button_cancel.clicked.connect(self.reject)
        self.button_login.clicked.connect(self.login)

        # Setup
        self.check_text()
        self.update_style_sheet()
        self.text_username.setFocus()

        main_layout.setSpacing(0)
        main_layout.setContentsMargins(0, 20, 0, 0)
        self.setLayout(main_layout)

    def setup(self):
        """Setup login dialog."""
        self.update_links()

    def update_links(self):
        """Update links."""
        for button in [
            self.button_forgot_username,
            self.button_forgot_password,
            self.button_register,
        ]:
            try:
                button.disconnect()
            except TypeError:  # pragma: no cover
                pass

        # TODO, get this from anaconda client directly?
        anaconda_api_url = self.config.get(
            'main', 'anaconda_api_url', self.anaconda_api.client_get_api_url()
        )
        if anaconda_api_url:
            # Remove api if using a subdomain
            base_url = anaconda_api_url.lower().replace('//api.', '//')
            self.base_url = base_url

            # Remove api if not using a subdomain
            parts = base_url.lower().split('/')
            if parts[-1] == 'api':
                base_url = '/'.join(parts[:-1])

            self.forgot_username_url = (
                base_url + '/' + self.FORGOT_USERNAME_URL
            )
            self.forgot_password_url = (
                base_url + '/' + self.FORGOT_PASSWORD_URL
            )

            self.button_register.clicked.connect(
                lambda: self.open_url(base_url)
            )
            self.button_forgot_username.clicked.connect(
                lambda: self.open_url(self.forgot_username_url)
            )
            self.button_forgot_password.clicked.connect(
                lambda: self.open_url(self.forgot_password_url)
            )

    def login(self):
        """
        Try to log the user in the specified anaconda api endpoint.
        If the 'anaconda_api_url' is not specified the error popup will arise.

        Otherwise, the login mechanism will be executed, by setting
        'logged_api_url' in config, and reloading the Client API.
        After that the login procedure will be run.
        """
        if not self.config.get('main', 'anaconda_api_url'):
            text = (
                'The Anaconda Cloud API domain is not specified! '
                'Please, set in preferences.'
            )
            msg_box = MessageBoxInformation(
                title='Login Error',
                text=text,
            )
            msg_box.exec_()

            self.button_login.setDisabled(False)
            self.check_text()
            QApplication.restoreOverrideCursor()

        else:
            self.button_login.setEnabled(False)
            self.text_username.setText(self.text_username.text().lower())
            QApplication.setOverrideCursor(Qt.WaitCursor)
            self.label_message.setText('')

            # Reload the client to the other one.
            anaconda_api_url = self.config.get('main', 'anaconda_api_url')
            self.config.set('main', 'logged_api_url', anaconda_api_url)
            self.anaconda_api.client_reload()

            worker = self.anaconda_api.login(
                self.text_username.text().lower(), self.text_password.text()
            )
            worker.sig_finished.connect(self._finished)

    def _finished(self, worker, output, error):
        """
        Callback for the login procedure after worker has finished.

        If success, sets the token, 'username' attribute to parent widget
        and sends the accept signal.

        Otherwise, outputs error messages.
        """
        token = output
        self._parent.username = self.username
        if token:
            self.token = token
            self.sig_authentication_succeeded.emit()
            self._parent.accept()
        elif error:
            username = self.text_username.text().lower()
            bold_username = '<b>{0}</b>'.format(username)

            # The error might come in (error_message, http_error) format
            try:
                error_message = ast.literal_eval(str(error))[0]
            except Exception:  # pragma: no cover
                error_message = str(error)

            error_message = error_message.lower().capitalize()
            error_message = error_message.split(', ')[0]
            error_text = '<i>{0}</i>'.format(error_message)
            error_text = error_text.replace(username, bold_username)
            self.label_message.setText(error_text)
            self.label_message.setVisible(True)

            if error_message:
                domain = self.anaconda_api.client_domain()
                label = '{0}/{1}: {2}'.format(
                    domain, username, error_message.lower()
                )
                self.tracker.track_event(
                    'authenticate',
                    'login failed',
                    label=label,
                )
                self.text_password.setFocus()
                self.text_password.selectAll()
            self.sig_authentication_failed.emit()

        self.button_login.setDisabled(False)
        self.check_text()
        QApplication.restoreOverrideCursor()

    def open_url(self, url):
        """Open given url in the default browser and log the action."""
        self.tracker.track_event('content', 'click', url)
        self.sig_url_clicked.emit(url)
        QDesktopServices.openUrl(QUrl(url))


class TeamEditionLoginTab(LoginTabMixin):
    def __init__(self, anaconda_api, parent, brand, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._brand = brand
        self._parent = parent
        self.config = CONF
        self.anaconda_api = anaconda_api
        self.team_edition_api = TeamEditionAPI(self.config.get('main', 'team_edition_api_url'), self.config)
        self.token = None
        self.tracker = GATracker()

        # Widgets
        self.label_username = QLabel('Username:')
        self.label_password = QLabel('Password:')
        self.text_username = QLineEdit()
        self.text_password = QLineEdit()

        self.label_information = QLabel(
            '''
            <strong>{brand}</strong> is where packages, notebooks,
            and <br> environments are shared between the teams. 
            <br> It provides powerful collaboration and package 
            <br>management for open source and private projects.<br>
            <hr>
            '''.format(brand=brand)
        )
        self.label_message = QLabel('')
        self.label_message.setWordWrap(True)

        self.button_login = ButtonPrimary('Login')
        self.button_cancel = ButtonNormal('Cancel')

        # Widgets setup
        self.button_login.setDefault(True)
        self.text_username.setValidator(QRegExpValidator(USER_RE))

        # This allows to completely style the dialog with css using the frame
        self.text_password.setEchoMode(QLineEdit.Password)
        self.label_message.setVisible(False)

        self.main_layout = QVBoxLayout()

        grid_layout = QVBoxLayout()
        grid_layout.addWidget(self.label_username)
        grid_layout.addWidget(self.text_username)
        grid_layout.addWidget(SpacerVertical())
        grid_layout.addWidget(self.label_password)
        grid_layout.addWidget(self.text_password)

        self.main_layout.addWidget(self.label_information)
        self.main_layout.addWidget(SpacerVertical())
        self.main_layout.addLayout(grid_layout)
        self.main_layout.addWidget(SpacerVertical())
        self.main_layout.addWidget(self.label_message)

        layout_buttons = QHBoxLayout()
        layout_buttons.addStretch()
        layout_buttons.addWidget(self.button_cancel)
        layout_buttons.addWidget(SpacerHorizontal())
        layout_buttons.addWidget(self.button_login)

        self.main_layout.addWidget(SpacerVertical())
        self.main_layout.addWidget(SpacerVertical())
        self.main_layout.addStretch()
        self.main_layout.addLayout(layout_buttons)

        self.main_layout.setSpacing(0)
        self.main_layout.setContentsMargins(0, 20, 0, 0)
        self.setLayout(self.main_layout)

        # Signals
        self.text_username.textEdited.connect(self.check_text)
        self.text_password.textEdited.connect(self.check_text)
        self.button_login.clicked.connect(self.login)
        self.button_cancel.clicked.connect(self.reject)

        # Setup
        self.check_text()
        self.update_style_sheet()
        self.text_username.setFocus()

    def login(self):
        """
        Try to log the user in the specified anaconda api endpoint.
        If the 'team_edition_api_url' is not specified the error popup will arise.

        Otherwise, the login mechanism will be executed, by setting
        'logged_api_url' in config, and reloading the Client API.
        After that the login procedure will be run.
        """
        if not self.config.get('main', 'team_edition_api_url'):
            text = (
                'The Team Edition API domain is not specified! '
                'Please, set in preferences.'
            )
            msg_box = MessageBoxInformation(
                title='Login Error',
                text=text,
            )
            msg_box.exec_()

            self.button_login.setDisabled(False)
            self.check_text()
            QApplication.restoreOverrideCursor()

        else:
            self.button_login.setEnabled(False)
            self.text_username.setText(self.text_username.text().lower())
            QApplication.setOverrideCursor(Qt.WaitCursor)
            self._parent.username = self.username

            # Reload the client to the other one.
            team_edition_api_url = self.config.get('main', 'team_edition_api_url')
            self.config.set('main', 'logged_api_url', team_edition_api_url)
            self.anaconda_api.client_reload()

            worker = self.anaconda_api.login(
                self.text_username.text().lower(), self.text_password.text()
            )
            worker.sig_finished.connect(self._finished)

    def _finished(self, worker, output, error):
        """
        Callback for the login procedure after worker has finished.

        If success, sets the token, 'username' attribute to parent widget
        and sends the accept signal.

        Otherwise, outputs error messages.
        """
        try:
            response = json.loads(output)

            if 'token' in response and 'refresh_token' in response:
                self._parent.username = self.username
                self.token = output
                self._parent.accept()
            else:
                if 'message' in response and response['message'] == 'Unauthorized':
                    self.label_message.setText('<i>Invalid Credentials!</i>')
                    self.label_message.setVisible(True)
                    self._track_error(output)

        except json.decoder.JSONDecodeError:
            self.label_message.setText('<i>Unhandled error happened!</i>')
            self.label_message.setVisible(True)
            self._track_error(error)

        finally:
            self.button_login.setDisabled(False)
            self.check_text()
            QApplication.restoreOverrideCursor()

    def _track_error(self, error):
        domain = self.anaconda_api.client_domain()
        label = '{0}/{1}: {2}'.format(
            domain, self.username, str(error).lower()
        )
        self.tracker.track_event(
            'authenticate',
            'login failed',
            label=label,
        )
        self.text_password.setFocus()
        self.text_password.selectAll()


class EnterpriseRepoLoginTab(LoginTabMixin):

    # See https://github.com/Anaconda-Platform/anaconda-server settings
    FORGOT_USERNAME_URL = 'account/forgot_username'
    FORGOT_PASSWORD_URL = 'account/forgot_password'

    sig_url_clicked = Signal(object)
    sig_authentication_succeeded = Signal()
    sig_authentication_failed = Signal()

    def __init__(self, anaconda_api, parent, brand, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._brand = brand
        self._parent = parent
        self.config = CONF
        self.anaconda_api = anaconda_api
        self.token = None
        self.error = None
        self.tracker = GATracker()
        self.forgot_username_url = None
        self.forgot_password_url = None

        # Widgets
        self.label_username = QLabel('Username:')
        self.label_password = QLabel('Password:')
        self.text_username = QLineEdit()
        self.text_password = QLineEdit()

        self.label_information = QLabel(
            '''
            <strong>{brand}</strong> is where packages, notebooks,
            and <br> environments are shared between the members of 
            <br>the enterprise teams. It provides powerful collaboration 
            <br> and package management for open source and private projects.<br>
            <hr>
            '''.format(brand=brand)
        )
        self.label_message = QLabel('')
        self.label_message.setWordWrap(True)

        self.button_login = ButtonPrimary('Login')
        self.button_cancel = ButtonNormal('Cancel')

        # Widgets setup
        self.button_login.setDefault(True)
        self.text_username.setValidator(QRegExpValidator(USER_RE))

        # This allows to completely style the dialog with css using the frame
        self.text_password.setEchoMode(QLineEdit.Password)
        self.label_message.setVisible(False)

        self.main_layout = QVBoxLayout()

        grid_layout = QVBoxLayout()
        grid_layout.addWidget(self.label_username)
        grid_layout.addWidget(self.text_username)
        grid_layout.addWidget(SpacerVertical())
        grid_layout.addWidget(self.label_password)
        grid_layout.addWidget(self.text_password)

        self.main_layout.addWidget(self.label_information)
        self.main_layout.addWidget(SpacerVertical())
        self.main_layout.addLayout(grid_layout)
        self.main_layout.addWidget(self.label_message)

        layout_buttons = QHBoxLayout()
        layout_buttons.addStretch()
        layout_buttons.addWidget(self.button_cancel)
        layout_buttons.addWidget(SpacerHorizontal())
        layout_buttons.addWidget(self.button_login)

        self.main_layout.addWidget(SpacerVertical())
        self.main_layout.addWidget(SpacerVertical())
        self.main_layout.addStretch()
        self.main_layout.addLayout(layout_buttons)

        # Signals
        self.text_username.textEdited.connect(self.check_text)
        self.text_password.textEdited.connect(self.check_text)
        self.button_cancel.clicked.connect(self.reject)
        self.button_login.clicked.connect(self.login)

        # Setup
        self.check_text()
        self.update_style_sheet()
        self.text_username.setFocus()

        self.main_layout.setSpacing(0)
        self.main_layout.setContentsMargins(0, 20, 0, 0)
        self.setLayout(self.main_layout)

    def login(self):
        """
        Try to log the user in the specified anaconda api endpoint.
        If the 'enterprise_4_repo_api_url' is not specified the error
        popup will arise.

        Otherwise, the login mechanism will be executed, by setting
        'logged_api_url' in config, and reloading the Client API.
        After that the login procedure will be run.
        """
        if not self.config.get('main', 'enterprise_4_repo_api_url'):
            text = (
                'The Enterprise 4 Repository API domain is not specified! '
                'Please, set in preferences.'
            )
            msg_box = MessageBoxInformation(
                title='Login Error',
                text=text,
            )
            msg_box.exec_()

            self.button_login.setDisabled(False)
            self.check_text()
            QApplication.restoreOverrideCursor()

        else:
            self.button_login.setEnabled(False)
            self.text_username.setText(self.text_username.text().lower())
            QApplication.setOverrideCursor(Qt.WaitCursor)
            self.label_message.setText('')

            # Reload the client to the other one, if needed.
            enterprise_4_repo_api_url = self.config.get('main', 'enterprise_4_repo_api_url')
            self.config.set('main', 'logged_api_url', enterprise_4_repo_api_url)
            self.anaconda_api.client_reload()

            worker = self.anaconda_api.login(
                self.text_username.text().lower(), self.text_password.text()
            )
            worker.sig_finished.connect(self._finished)

    def _finished(self, worker, output, error):
        """
        Callback for the login procedure after worker has finished.

        If success, sets the token, 'username' attribute to parent widget
        and sends the accept signal.

        Otherwise, outputs error messages.
        """
        token = output
        self._parent.username = self.username
        if token:
            self.token = token
            self.sig_authentication_succeeded.emit()
            self._parent.accept()
        elif error:
            username = self.text_username.text().lower()
            bold_username = '<b>{0}</b>'.format(username)

            # The error might come in (error_message, http_error) format
            try:
                error_message = ast.literal_eval(str(error))[0]
            except Exception:  # pragma: no cover
                error_message = str(error)

            error_message = error_message.lower().capitalize()
            error_message = error_message.split(', ')[0]
            error_text = '<i>{0}</i>'.format(error_message)
            error_text = error_text.replace(username, bold_username)
            self.label_message.setText(error_text)
            self.label_message.setVisible(True)

            if error_message:
                domain = self.anaconda_api.client_domain()
                label = '{0}/{1}: {2}'.format(
                    domain, username, error_message.lower()
                )
                self.tracker.track_event(
                    'authenticate',
                    'login failed',
                    label=label,
                )
                self.text_password.setFocus()
                self.text_password.selectAll()
            self.sig_authentication_failed.emit()

        self.button_login.setDisabled(False)
        self.check_text()
        QApplication.restoreOverrideCursor()


class AuthenticationDialog(DialogBase):
    """Login dialog."""

    def __init__(self, anaconda_api, parent=None, brand=DEFAULT_BRAND):
        """Login dialog."""
        super(AuthenticationDialog, self).__init__(parent)

        self._username = ''

        self.setMinimumWidth(260)
        self.setWindowTitle('Sign in')

        self.cloud_tab = AnacondaCloudLoginTab(anaconda_api, self, brand)
        self.team_edition_tab = TeamEditionLoginTab(anaconda_api, self, TEAM_EDITION_BRAND)
        self.enterprise_repo4_tab = EnterpriseRepoLoginTab(anaconda_api, self, ENTERPRISE_BRAND)

        self.tab_widget = QTabWidget()
        self.tab_widget.addTab(self.cloud_tab, 'Anaconda Cloud')
        self.tab_widget.addTab(self.team_edition_tab, 'Team Edition')
        self.tab_widget.addTab(self.enterprise_repo4_tab, 'Enterprise 4')

        main_layout = QVBoxLayout()
        main_layout.addWidget(self.tab_widget)

        self.setLayout(main_layout)
        self.setup()

    def setup(self):
        self.cloud_tab.setup()

    @property
    def username(self):
        return self._username

    @username.setter
    def username(self, val):
        self._username = val


# --- Local testing
# -----------------------------------------------------------------------------
def local_test():  # pragma: no cover
    """Run local test."""
    from anaconda_navigator.utils.qthelpers import qapplication
    app = qapplication(test_time=3)
    load_fonts(app)
    widget = AuthenticationDialog(AnacondaAPI())
    widget.update_style_sheet()
    widget.show()
    app.exec_()


if __name__ == '__main__':  # pragma: no cover
    local_test()
